; (c) 1992 Dirk Schwarzhans, Matthias Braun
; Use under the terms of the GNU General Public License
   

  include 'ioequ.asm'

; Konstanten
DMA_SIZE		equ	$0400	; Gre jedes DMA-Blocks = 1024 Bytes
DMA_BUFFER		equ	$3C00	; Pufferadresse fr DMA
CMD_SIZE		equ	$200	; Gre des Kommandopuffers
CMD_BUFFER		equ	$3A00	; Pufferadresse fr Kommandos

; Messages DSP --> Host
DM_R_REQ		equ	$050001	; DMA-Anforderung Lesen
DM_PAGE_1		equ	$100000	; Seitendaten Teil 1
DM_PAGE_2		equ	$110000	; Seitendaten Teil 2
DM_PAGE_3		equ	$120000	; Seitendaten Teil 3
DM_CLK_1		equ	$130000	; Uhrzeit Teil 1
DM_CLK_2		equ	$140000	; Uhrzeit Teil 2
DM_CLK_3		equ	$150000	; Uhrzeit Teil 3
DM_CLK_4		equ	$160000	; Uhrzeit Teil 4
DM_STOP			equ	$170000	; DSP hat angehalten

; Message Host --> DSP
HM_REQ_PAGE		equ	$200000	; setzt Anforderungsdaten
HM_DISP_CNTL	equ	$210000	; setzt die Display-Control-Register
HM_DISP_MODE	equ	$220000	; setzt das Display-Modus-Register
HM_DSP_STOP		equ	$230000	; Arbeit einstellen
HM_SYS_CONFIG	equ	$240000	; Messages und Uhrzeit schalten
HM_MODE_REG		equ	$250000	; setzt die untersten 3 Bits des Mode-Registers
MIN_HM_ID		equ	$20		; Befehlsbyte der kleinsten HM
NUM_HM			equ	6		; Anzahl der gltigen Host-Messages

; Host-Kommandos
VEC_R_DONE		equ	$0024	; DMA: Lesen fertig
VEC_HRI			equ	$0020	; Host-Empfangsregister voll

; IIC-Leitungen
SCL_OUT			equ $5		; Bit-Nummer der Taktausgabeleitung
SDA_OUT			equ $4		; Bit-Nummer der Datenausgabeleitung
SDA_IN  		equ $3		; Bit-Nummer der Dateneingabeleitung
PORT_C_DIR		equ (1<<SCL_OUT)|(1<<SDA_OUT)
PORT_C_INIT		equ (1<<SCL_OUT)
READAD			equ $23		; Adresse des IVT zum Lesen
WRITEAD			equ $22		; Adresse des IVT zum Schreiben

; ---------------------------------------------------------------

; Variablen
  org	y:0

; Daten ber die Benutzung der Kanle
B_ACTIVE		equ	0		; Kanal wartet auf Kopfzeile (nur Kanal 0)
B_FREE			equ	1		; Kanal frei (nur Kanle 1-3)
B_READY			equ	2		; Kanal auslesebereit (nur Kanle 0,2,3)
circuit_0_stat	dc	0
circuit_1_stat	dc	(1<<B_FREE)
circuit_2_stat	dc	(1<<B_FREE)
circuit_3_stat	dc	(1<<B_FREE)

; wir mssen uns merken, welcher Kanal die letzte Kopfzeile bekommen hat
last_header		dc	1		; zunchst ungltig

; auerdem ist es praktisch sich hier getrennt zu merken, welcher der oberen
; drei Kanle gerade auf eine Kopfzeile wartet
active_circuit	dc	3

; Hier werden die Anforderungsdaten gespeichert
circuit_0_mag	dc	1|%11000	; nach Start Seite 100 suchen
circuit_0_p10	dc	0|%10000
circuit_0_pu	dc	0|%10000

; System-Zustandswort; Bedeutung der Bits:
B_MSG_ON		equ	0		; Messages fr angekommene Seiten an
B_CLK_ON		equ	1		; Uhrzeitmessages eingeschaltet
B_PAGE_VALID	equ	2		; erwarteter Pageheader angekommen
sys_stat		dc	0		; zunchst ist alles ausgeschaltet

; DMA-Status-Wort: Momentaner Zustand der bertragung; Bits:
B_DMA_DONE		equ	0		; DMA fertig
DMA_stat		dc	0

; Tabelle mit den Anfangsadressen der einzelnen Blcke
row_col_tab		dc	0,0
				dc	6,0
				dc	12,0
				dc	18,0
block_count		equ	4		; insgesamt vier Blcke
block_size		equ	240		; a 240 Zeichen

; Tabellen mit entsprechend vorbereiteten Messages
clock_table		dc	DM_CLK_1>>16
				dc	DM_CLK_2>>16
				dc	DM_CLK_3>>16
				dc	DM_CLK_4>>16
page_table		dc	DM_PAGE_1>>15
				dc	DM_PAGE_2>>15
				dc	DM_PAGE_3>>15

; Speicherplatz fr die Seitendaten
circuit_status	dc	0,0,0,0,0,0,0,0,0,0

; Speicherplatz fr die alte und neue Uhrzeit
old_clock		dc	0,0,0,0,0,0,0,0
new_clock		dc	0,0,0,0,0,0,0,0

; Speicherplatz fr Register bei Interrupts
saved_a2		dc	0
saved_a1		dc	0
saved_a0		dc	0
saved_x1		dc	0

; die Sprungtabelle fr die Host-Messages steht im x-Speicher
  org	x:0 ; mu bei der jetzigen Implementierung Adresse 0 sein
jumptable		dc	req_page_handler
				dc	disp_cntl_handler
				dc	disp_mode_handler
				dc	dsp_stop_handler
				dc	sys_config_handler
				dc	mode_handler

; ---------------------------------------------------------------

; Macros
			
; ein Byte ber den IIC-Bus senden; zerstrt den Inhalt von a
SENDBYTE	macro	data
			move	#data,a1	; bei move nach a1 kein '>' ntig
			jsr		<sendbyte
			endm

; sendet a ber den IIC-Bus; zerstrt den Inhalt von a
SENDA		macro
			jsr		<sendbyte
			endm

; beginnt eine Schreibsequenz beginnend mit der angegebenen
; Registeradresse; zerstrt den Inhalt von a
STARTWRITE	macro		register
			jsr			<startiic
			SENDBYTE	WRITEAD
			SENDBYTE	register
			endm

; beendet eine Schreibsequenz
STOPWRITE	macro	
			jsr			<stopiic
			endm

; beginnt eine Lesesequenz; zerstrt den Inhalt von a
STARTREAD	macro
			jsr			<startiic
			SENDBYTE	READAD
			endm

; beendet eine Lesesequenz
STOPREAD	macro	
			jsr			<stopiic
			endm

; ---------------------------------------------------------------

; Interruptvektoren
  org	p:$0
  jmp	>reset

  org	p:VEC_HRI
  jsr	>HM_int

  org	p:VEC_R_DONE
  bset	#B_DMA_DONE,y:DMA_stat
  nop

; Initialisierung
  org	p:$40
reset
  movec	#6,omr				; data rom enabled, mode 2
  movec	#0,sp				; Stack lschen
  bset	#0,x:m_pbc			; host port einschalten
  movep	#>$000000,x:m_bcr	; keine Wartezyklen auf externem RAM
  movep	#>$000400,x:m_ipr  	; Intr levels: HOST=0, andere aus
  bset	#m_hcie,x:m_hcr		; host command Interrupts erlauben
  move	#CMD_BUFFER,r7		; HM-Schreibzeiger
  move	#(CMD_SIZE-1),m7
  move	r7,r6				; und HM-Lesezeiger initialisieren
  move	m7,m6
  bset	#m_hrie,x:m_hcr		; host recieve Interrupts erlauben
  move	#3,sr				; niedrigste Interruptmaske im MR
  movep	#>$000000,x:m_pcc	; Port C als paralleles Interface
  movep	#>PORT_C_DIR,x:m_pcddr	; Richtungen von Port C setzen

  movep	#>PORT_C_INIT,x:m_pcd	; SCA = SCL = high	
  rep	#0					; einige ms verzgern
  nop

  clr	a	#DMA_BUFFER,r0	; Puffer lschen, damit man beliebig
  rep	#DMA_SIZE			; viele Zeilen aus dem Chip lesen
  move	a,y:(r0)+			; kann, ohne viel zu ndern

  jmp	extram

; ---------------------------------------------------------------

; Interruptroutinen

; empfngt eine Host-Message und fgt sie in die Queue ein; ist diese
; voll, so wird die lteste berschrieben
HM_int
  move	x1,y:saved_x1
  move	a2,y:saved_a2
  move	a1,y:saved_a1
  move	a0,y:saved_a0
  lua	(r7)+,n7
  move	r6,x1
  move	n7,a
  cmp	x1,a
  jne	<_not_full
  lua	(r6)+,r6	; ltestes Kommando berschreiben
_not_full
  movep	x:m_hrx,a
  move	a,y:(r7)+
  move	y:saved_a0,a0
  move	y:saved_a1,a1
  move	y:saved_a2,a2
  move	y:saved_x1,x1
  rti

; allgemeine Unterprogramme

; liest 'block_size' Bytes aus Kanal x0 in den Seitenpuffer 
; r1 zeigt auf Zeile und Spalte und wird entsprechend erhht
read_ivt_block
  STARTWRITE	8			; Register 8 = aktiver Kanal
  move	x0,a				; gewnschte Kanalnummer senden
  SENDA
  move	y:(r1)+,a			; Zeile...
  SENDA
  move	y:(r1)+,a			;        und Spalte programmieren
  SENDA
  STOPWRITE
  STARTREAD
 .LOOP	#(block_size-1)
  jsr	<receive			; Daten abholen
  move	a1,y:(r0)+			; und in den Puffer schreiben
 .ENDL
  jsr	<receivelast		; letztes Byte abholen
  move	a1,y:(r0)+
  STOPREAD
  rts

; liest die Status-Informationen von Kanal x0 in den Seitenpuffer
read_ivt_status
  move	#(DMA_BUFFER+1000),r0
  STARTWRITE 	8
  move	x0,a
  SENDA
  SENDBYTE	25				; Zeile 25
  SENDBYTE	0				; Spalte 0
  STOPWRITE
  STARTREAD
 .LOOP	#9
  jsr	<receive			; Daten abholen
  move	a1,y:(r0)+			; und in den Puffer schreiben
 .ENDL
  jsr	<receivelast		; letztes Byte abholen
  move	a1,y:(r0)+
  STOPREAD
  rts

; liest die Daten aus Kanal x0 in den Seitenpuffer
; verndert a, x1, y0, r0, r1, r2, n2
read_data
  move	#row_col_tab,r1
  move	#DMA_BUFFER,r0
 .LOOP	#block_count
  jsr	read_ivt_block		; einen Block lesen
  jsr	test_circuits		; testen, ob neuer Header angekommen
  nop						; erforderlich wegen der do-Schleife
 .ENDL						; die Seitendaten sind jetzt komplett
  jsr	read_ivt_status		; Statusinformationen hinzufgen
  rts

; liest die Uhrzeit aus Kanal 0 in den Seitenpuffer
; verndert a, r0
read_clock
  STARTWRITE 	8
  SENDBYTE	0				; Kanal 0
  SENDBYTE	0				; Zeile 0
  SENDBYTE	32				; Spalte 32
  STOPWRITE
  move	#(DMA_BUFFER+32),r0
  STARTREAD
 .LOOP	#7
  jsr	<receive			; Daten abholen
  move	a1,y:(r0)+			; und in den Puffer schreiben
 .ENDL
  jsr	<receivelast		; letztes Byte abholen
  move	a1,y:(r0)+
  STOPREAD
  rts

; liest die Uhrzeit aus Kanal 0 in den Uhrzeit-Puffer new_clock
; benutzt a und r2
read_ivt_clock
  STARTWRITE	8
  SENDBYTE	0				; Kanal 0
  SENDBYTE	0				; Zeile 0
  SENDBYTE	32				; Spalte 32
  STOPWRITE
  move	#new_clock,r2
  STARTREAD
 .LOOP	#7
  jsr	<receive			; Daten abholen
  move	a1,y:(r2)+			; und in den Puffer schreiben
 .ENDL
  jsr	<receivelast		; letztes Byte abholen
  move	a1,y:(r2)+
  STOPREAD
  rts

; liest die Statusinformationen von Kanal x1 in y:circuit_status
; benutzt a, y0 und r2
read_test_status
  STARTWRITE	8
  move	x1,a				; Kanal
  SENDA
  SENDBYTE	25				; Zeile 25
  SENDBYTE	9				; Spalte 9 fr PBLF
  STOPWRITE
  STARTREAD
  jsr	<receivelast		; letztes Byte abholen
  move	a1,y:(circuit_status+9)
  STOPREAD
  move	#>$1f,y0			; Maske fr die ersten 9 Bytes
  STARTWRITE	8
  move	x1,a				; Kanal
  SENDA
  SENDBYTE	25				; Zeile 25
  SENDBYTE	0				; Spalte 0
  STOPWRITE
  move	#circuit_status,r2
  STARTREAD
 .LOOP	#8
  jsr	<receive			; Daten abholen
  and	y0,a				; maskieren falls bertragungsfehler
  move	a1,y:(r2)+			; und in den Puffer schreiben
 .ENDL
  jsr	<receivelast		; letztes Byte abholen
  and	y0,a				; maskieren falls bertragungsfehler
  move	a1,y:(r2)+
  STOPREAD
no_stat_read
  rts

; sendet den Puffer per "DSP-initiated DMA"; blockiert dabei, 
; bis der gesamte Puffer bertragen ist; verwendet r2
dma_send
  move	#DMA_BUFFER,r2		
  bclr	#B_DMA_DONE,y:DMA_stat	; DMA-Ende-Bit zuruecksetzen
  jclr	#m_htde,x:m_hsr,*	; warten, bis wir senden drfen
  movep	#DM_R_REQ,x:m_htx	; DMA DSP->Host anfordern
  jclr	#m_hf1,x:m_hsr,*	; warten, bis HF1 = 1 (DMA-Start)
  do	#DMA_SIZE,_dmaLoop	; eigentliche Sendeschleife
  jclr	#m_htde,x:m_hsr,*	; warten, bis senden erlaubt
  movep	y:(r2)+,x:m_htx		; und ab geht die Post
_dmaLoop
  jset	#B_DMA_DONE,y:DMA_stat,_endDMA	; Fertig? Dann raus hier!
  jclr  #m_htde,x:m_hsr,_dmaLoop	; sonst: Knnen wir senden?
  movep #0,x:m_htx					; dann Nullen senden
  jmp	_dmaLoop
_endDMA
  jset	#m_hf1,x:m_hsr,*	; warten, bis DMA beendet
  rts

; programmiert den Kanal 0 mit den vorliegenden Daten
; verwendet r2,a
start_circuit_0
  STARTWRITE	2
  SENDBYTE	0				; Kanal 0
  move	#circuit_0_mag,r2
 .LOOP	#3
  move	y:(r2)+,a			; Anforderungsdaten programmieren
  SENDA
  nop						; wegen der Schleife
 .ENDL	
 .LOOP	#4
  SENDBYTE	0				; nur der Vollstndigkeit halber
  nop
 .ENDL						; Unterseiten = don't care
  STOPWRITE
  rep	#1000				; kurz warten, bis PBLF gesetzt ist
  nop
  bset	#B_ACTIVE,y:circuit_0_stat
  rts

; hlt den Empfang auf Kanal 0 an; verwendet a
stop_circuit_0
  STARTWRITE	2
  SENDBYTE	0				; Kanal 0
  SENDBYTE	0				; hold aktivieren
  STOPWRITE
  rts

; lscht das found-Bit von Kanal 0; verndert a
clear_found_0
  STARTWRITE	8
  SENDBYTE	0
  SENDBYTE	25
  SENDBYTE	8
  SENDBYTE	16				; found = 0
  STOPWRITE
  rts

; hlt Kanal 1 an und markiert ihn als frei
; verndert a
stop_circuit_1
  STARTWRITE	2
  SENDBYTE	$10				; Nummer des Kanals
  SENDBYTE	0				; hold aktivieren
  STOPWRITE
  bset	#B_FREE,y:circuit_1_stat
  rts

; testet bei eingeschalteten Uhren-Message auf vernderte Uhrzeit
; und sendet evtl. die entsprechenden Messages
; benutzt a, b, r2, r3, x0
check_clock
  jclr	#B_CLK_ON,y:sys_stat,no_clock	; Uhren-Messages sind aus
  jsr	<read_ivt_clock		; neue Uhrzeit vom IVT holen
  clr	b	#old_clock,r2	; b dient als Merker, ob
  move	#new_clock,r3		; sich die Zeit verndert hat
 .LOOP	#8
  move	y:(r2),a			; Zeichen aus alter Uhrzeit -> a
  move	y:(r3)+,x0			; Zeichen aus neuer Uhrzeit -> x0
  cmp	x0,a	x0,y:(r2)+	; vergleichen; neue Zeit bernehmen
  jeq	_no_change
  bset	#0,b1				; Vernderung: Flag setzen
_no_change
  nop						; wegen der Schleife
 .ENDL
  jclr	#0,b1,no_clock		; keine Vernderung
  move	#clock_table,r2
  move	#new_clock,r3
 .LOOP	#4					; 4 Messages senden
  move	y:(r2)+,a			; entsprechenden Header holen
 .LOOP	#2					; 2 Bytes pro Message
  rep	#8					; Message 8 Bits nach links schieben
  asl	a
  move	y:(r3)+,x0			; Uhren-Byte holen
  or	x0,a				; und in die Message einfgen
 .ENDL
  jclr	#m_htde,x:m_hsr,*	; warten, bis wir senden drfen
  movep	a,x:m_htx			; und Message abschicken
 .ENDL
no_clock
  rts

; sendet bei eingeschalteten Seiten-Messages eine solche
; verwendet a, y0, r2, r3
send_page_message
  jclr	#B_MSG_ON,y:sys_stat,no_page
  move	#page_table,r2
  move	#circuit_status,r3
 .LOOP	#3					; 3 Messages senden
  move	y:(r2)+,a			; Header holen
 .LOOP	#3					; 3 Infos pro Message
  rep	#5					; Message 5 Bits nach links schieben
  asl	a
  move	y:(r3)+,y0			; Seiteninformation holen
  or	y0,a				; und einfgen
 .ENDL
  jclr	#m_htde,x:m_hsr,*	; warten, bis wir senden drfen
  movep	a,x:m_htx			; und Message abschicken
 .ENDL
no_page
  rts

	include		'IIC.asm'

; ---------------------------------------------------------------

; Programm im externen Speicher

  org	p:$2000
extram

; Initialisierung des VT-Chips
  STARTWRITE	0
  SENDBYTE	0				; advanced control
  SENDBYTE	0				; mode
  STOPWRITE

; alle Kanle abschalten
  clr	b	#>$10,x0		; Start mit Kanal 0
 .LOOP	#4
  STARTWRITE	2
  move	b,a					; Kanalnummer senden
  SENDA
 .LOOP		#7
  SENDBYTE	0				; don't care
  nop						; wegen der Schleife
 .ENDL
  STOPWRITE
  add	x0,b				; nchster Kanal
 .ENDL
  STARTWRITE	4			; Display-Daten setzen
  SENDBYTE	0				; Display-Kapitel = 0
  SENDBYTE	%11001100		; Text und Hintergrund einschalten
  SENDBYTE	%00010111		; R6
  SENDBYTE	%00000111		; Display mode
  STOPWRITE
  jsr	start_circuit_0	; Kanal 0 einschalten
  move	#3,n2
  jsr	start_next_free_circuit
  
main_loop
  move	r6,a
  move	r7,x0
  cmp	x0,a
  jeq	_no_more_HM			; keine HMs mehr verfgbar
  move	y:(r6)+,x0			; HM -> x0
  tfr	x0,a	#>$80,y0	; HM -> a, Faktor fr Shift -> y0
  mpy	x0,y0,b	#>MIN_HM_ID,x0	; HM-ID-> b; min. HM -> x0
  sub	x0,b	#>NUM_HM,x0	; Index in Sprungtabelle -> b
  jcs	main_loop			; ungltige HM
  cmp	x0,b	b,r1		; Adresse der Sprungadresse -> r1
  jcc	main_loop			; ungltige Host-Message
  move	x:(r1),r1			; Sprungvektor -> r1
  nop						; warten, bis r1 verfgbar
  jsr	(r1)				; HM-Handler anspringen
  jmp	main_loop			; nchste HM auswerten

_no_more_HM
  jsr	check_clock			; evtl. Uhrzeit an Host senden

  jsr	test_circuits		; irgendetwas angekommen?
  jclr	#B_READY,y:circuit_0_stat,_no_data_0
  move	#>0,x0
  jsr	read_data			; Daten auslesen
  jsr	read_ivt_status		; Statusinformationen hinzufgen
  jsr	dma_send			; und die Daten an den Host senden
  bclr	#B_READY,y:circuit_0_stat
  jsr	start_circuit_0
  jmp	main_loop

; die folgenden beiden Teile knnte man in sicherlich ausfaktorisieren...
_no_data_0  
  jclr	#B_READY,y:circuit_3_stat,_no_data_3
  move	#>3,x0
  jsr	read_data			; Daten auslesen
  jsr	read_ivt_status		; Statusinformationen hinzufgen
  jsr	read_clock			; ... Uhrzeit dazu...
  jsr	dma_send			; und die Daten an den Host senden
  bclr	#B_READY,y:circuit_3_stat
  bset	#B_FREE,y:circuit_3_stat
  jclr	#B_READY,y:circuit_2_stat,_no_start_3
  jsr	stop_circuit_1
  STARTWRITE	2
  SENDBYTE	$30				; Nummer des Kanals
  SENDBYTE	$08				; hold abschalten
  STOPWRITE
  rep	#1000
  nop
  bclr	#B_FREE,y:circuit_3_stat
  move	x0,y:active_circuit
_no_start_3 
  jmp	main_loop

_no_data_3 
  jclr	#B_READY,y:circuit_2_stat,main_loop
  move	#>2,x0
  jsr	read_data			; Daten auslesen
  jsr	read_ivt_status		; Statusinformationen hinzufgen
  jsr	read_clock			; ... Uhrzeit dazu...
  jsr	dma_send			; und die Daten an den Host senden
  bclr	#B_READY,y:circuit_2_stat
  bset	#B_FREE,y:circuit_2_stat
  jclr	#B_READY,y:circuit_3_stat,_no_start_2
  jsr	stop_circuit_1
  STARTWRITE	2
  SENDBYTE	$20				; Nummer des Kanals
  SENDBYTE	$08				; hold abschalten
  STOPWRITE
  rep	#1000
  nop
  bclr	#B_FREE,y:circuit_2_stat
  move	x0,y:active_circuit
_no_start_2 
  jmp	main_loop


; Testen aller wartenden Kanle
; verndert a, x1, y0, r2, n2
test_circuits
  move	y:active_circuit,x1
  jsr	read_test_status
  jset	#5,y:(circuit_status+9),_no_random_page
  jsr	process_last_channel
  move	x1,y:last_header
  move	#>1,a
  cmp	x1,a
 .IF <EQ>					; Kopfzeile von Auffangkanal empfangen
  jsr	stop_circuit_1		; Auffangkanal ausschalten und als frei markieren
  move	#3,n2
  jsr	start_next_free_circuit
 .ELSE
  move	x1,n2
  jsr	start_next_free_circuit
 .ENDI
  jsr	send_page_message
_no_random_page
  jclr	#B_ACTIVE,y:circuit_0_stat,_return
  move	#>0,x1
  jsr	read_test_status
  jset	#5,y:(circuit_status+9),_return
  bclr	#B_ACTIVE,y:circuit_0_stat
  jsr	process_last_channel
  move	x1,y:last_header
  jsr	send_page_message
_return
  rts

; wenn letzter Kanal gltig, dann ausschalten und als auslesebereit markieren
; zerstrt a, y0, r2, n2
process_last_channel
  move	y:last_header,y0
  move	#>1,a				; 1 ist der Auffangkanal
  cmp	y0,a
 .IF <NE>					; letzter Kanal gltig
  move	#circuit_0_stat,r2
  move	y0,n2
  clr	a
  cmp	y0,a				; letzter Kanal = Kanal 0?
 .IF <EQ>
  jsr	stop_circuit_0		; evtl. auch nichts machen
 .ELSE
  STARTWRITE	2
  move	y0,a
  rep	#4
  asl	a
  SENDA						; Nummer des Kanals
  SENDBYTE	0				; hold aktivieren
  STOPWRITE
 .ENDI
  bset	#B_READY,y:(r2+n2)	; Kanal als auslesebereit markieren
 .ENDI
  rts

; startet den hchsten freien Zufallskanal kleiner oder gleich n2
; verndert a, y0, r2, n2
start_next_free_circuit
  move	#circuit_0_stat,r2
  move	#>1,y0
_loop
  jset	#B_FREE,y:(r2+n2),_found
  move	n2,a
  sub	y0,a
  move	a,n2
  jmp	_loop
_found
  bclr	#B_FREE,y:(r2+n2)
  STARTWRITE	2
  move	n2,a
  move	a,y:active_circuit
  rep	#4
  asl	a
  SENDA						; Kanal
  SENDBYTE	$08				; hold ausschalten
  STOPWRITE
  rep	#1000				; kurz warten, bis PBLF gesetzt ist
  nop
  rts

; Host-Message Unterprogramme

req_page_handler
  move	#>$f,y0				; Maske -> y0
  move	#circuit_0_pu,r2
  bset	#11,a1				; hold ausschalten
 .LOOP	#3
  tfr	a,b
  and	y0,b				; gewnschte Daten extrahieren
  bset	#4,b1				; do-care-Bits setzen
  move	b,y:(r2)-			; und in den Speicher schreiben
  rep	#4					; nchste Daten zurechtschieben
  asr	a
 .ENDL
  jclr	#B_ACTIVE,y:circuit_0_stat,_return
  jsr	stop_circuit_0		; wir mssen Kanal 0 ausschalten
  jsr	start_circuit_0		; und wieder starten
_return
  rts

disp_cntl_handler
  tfr	a,b					; Message retten
  STARTWRITE	5			; normal display control
  tfr	b,a
  rep	#8					; normal display control extrahieren
  asr	a
  SENDA
  tfr	b,a
  SENDA
  STOPWRITE
  rts

disp_mode_handler
  tfr	a,b					; Message retten
  STARTWRITE	7			; display mode
  tfr	b,a
  SENDA
  STOPWRITE
  rts

dsp_stop_handler
  jclr	#m_htde,x:m_hsr,*	; warten, bis wir senden drfen
  movep	#>DM_STOP,x:m_htx	; und Stop-Message abschicken
  jmp	*					; ENDE
  
sys_config_handler
  bclr	#B_MSG_ON,y:sys_stat
  bclr	#B_CLK_ON,y:sys_stat
  jclr	#0,a1,page_messages_off
  bset	#B_MSG_ON,y:sys_stat
page_messages_off
  jclr	#1,a1,clock_messages_off
  bset	#B_CLK_ON,y:sys_stat
clock_messages_off
  rts

mode_handler
  tfr	a,b		#>7,y0		; Message retten
  and	y0,b				; Bits maskieren
  STARTWRITE	1			; display mode
  tfr	b,a
  SENDA
  STOPWRITE
  rts

